from . import rule
from .field import Field, StringField, BoolField, EnumField, IntField, DateField, BigIntField
from .consts import frequencies, id_string, calculated_string, frequency_string, min_value_string, max_value_string, \
    start_time_string, is_integer_string, delay_string, chunks_string


class TimeseriesMetadataField(Field):
    """
        Company Domain - I think we should take this from mainDomain. Each company should have one
        Ticker - do we use ticker/exchange directly to pull data? I thought we use EOD?
        Exchange - do we use ticker/exchange directly to pull data? I thought we use EOD?
        EOD ticker - should be exactly one. Flag cases of missing and multiple. All companies should have.
        Linkedin URL - should be exactly one, or in rare cases missing. Flag cases of missing and multiple. MOST companies should have.
        Indeed URL - should be exactly one, or in rare cases missing. Flag cases of missing and multiple.
        Comparably URL - should be exactly one, but will be missing for MANY companies. Flag cases of missing and multiple.
        Marketbeat URL - should be exactly one, or in rare cases missing. Flag cases of missing and multiple. MOST companies should have.
        Yahoo Finance URL - Is this for ESG? So only relevant for s&p500? or we need it for other things as well?
        Twitter - not too sure actually. I think should be one.
        Facebook - not too sure actually. I think should be one.
    """
    def __init__(self):
        super().__init__()
        self.name = "timeseries"
        self.value = {}

        self.id = StringField(id_string)
        self.calculated = BoolField(calculated_string)
        self.frequency = EnumField(frequency_string, frequencies)  # Check for daily, etc. more than/ less than daily -> gaps
        self.chunks = EnumField(chunks_string, frequencies)  # Used for recent data
        self.min_value = IntField(min_value_string)  # done
        self.max_value = IntField(max_value_string)  # done
        self.start_time = DateField(start_time_string)  # done
        self.is_integer = BoolField(is_integer_string)  # done
        self.delay = BigIntField(delay_string)  # goes w/ frequency. If delay =week and frequency=daily then last value should be a week ago.
        self.fields = [self.id, self.calculated, self.frequency, self.min_value, self.max_value,
                       self.start_time, self.is_integer, self.delay]
        self.rules = [rule.has_field(x.name) for x in self.fields]

    def set(self, value):
        self.value = value
        self.run_rules()
        if self.errors:  # checks for field names in value (type: dict)
            return self.errors
        self.set_fields(self.fields)
        return self.errors



from .field import Field
from .data_object import DataObject


class DataList(Field):
    def __init__(self):
        super().__init__()
        self.name = "data"
        self.value = []  # list of TimeseriesObjectField
        self.rules = []

    def append_obj(self, lst):
        observation = DataObject()
        err = observation.set(lst) # checks length = 2 then sets
        if err:
            self.error_many(err)
            return
        self.value.append(observation)
        return self.errors

    def set(self, value):
        for lst in value:
            self.append_obj(lst)
        return self.errors

    def validate(self) -> list:
        for field in self.value:
            self.validate_field(field)
        if self.errors:
            return self.errors
        self.run_rules()
        if not self.errors:
            self.valid = True
        return self.errors



import logging
from datetime import datetime
from . import rule
from .timeseries_response import TimeseriesResponseField
from .field import DateField


class DateCheck(TimeseriesResponseField):
    def __init__(self):
        """
        :param metadata_fields: list of string names
        """
        super().__init__()
        self.name = "date checks"
        self.metadata.fields = [
            self.metadata.start_time,
            self.metadata.delay
        ]
        self.metadata.rules = [
            rule.has_field(self.metadata.start_time.name),
            rule.has_field(self.metadata.delay.name)
        ]
        self.data.rules = [
            rule.Rule("Check has valid date after start_time", lambda x: self.valid_date(x)),
            rule.Rule("Check has recent date", lambda x: self.recent_date(x)),
        ]

    def valid_date(self, data_lst):
        # running check on list of TimeseriesObjectField
        start_time = self.metadata.start_time
        for obj in data_lst:
            if obj.date.before(start_time):
                logging.error("TimeseriesDateCheck: date {} is before start_time {}"
                              .format(obj.date.value, start_time.value))
                return False
        return True

    def recent_date(self, data_lst):
        recent = DateField("now")
        recent.value = datetime.now() - datetime.fromtimestamp(self.metadata.delay // 1e9)
        for date_obj in data_lst[::-1]:
            if date_obj.after(recent):
                return True
        return False



import logging

from datamodelz.error import Error


class Check:
    """
    Takes a list of DataObject and runs the function on the entire list. Returns an error string.
    """
    def __init__(self, name: str, funct) -> None:
        self.name = name
        self.funct = funct

    def run(self, data_lst, metadata=None, other_data=None) -> str:
        logging.debug("running check {}".format(self.name))
        return self.funct(data_lst, metadata, other_data)


class CheckEach(Check):
    """
    Takes a list of DataObject and runs the function on each DataObject within the list. Returns an error string.
    """
    def __init__(self, name: str, funct) -> None:
        super().__init__(name, funct)

    def run(self, data_lst, metadata=None, other_data=None) -> str:
        logging.debug("running check all {}".format(self.name))
        for obj in data_lst:
            error = self.funct(obj, metadata, other_data)
            if error:
                return error
        return Error()

class BusinessCheck(Check):
    def __init__(self, name: str, funct) -> None:
        super().__init__(name, funct)

    def run(self, data_lst) -> Error:
        logging.debug("running check {}".format(self.name))
        return self.funct(data_lst)

class BusinessCheckEach(Check):
    """
    Takes a list of DataObject and runs the function on each DataObject within the list. Returns an error string.
    """
    def __init__(self, name: str, funct) -> None:
        super().__init__(name, funct)

    def run(self, data_lst) -> Error:
        logging.debug("running check each {}".format(self.name))
        for obj in data_lst:
            error = self.funct(obj)
            if error:
                return error
        return Error()



from datetime import datetime


class Error:
    def __init__(self,
                 timeseries_code: str = "",
                 company: str = "",
                 check_name: str = "",
                 date: datetime = None,
                 value=None,
                 business_rule=None,
                 reference: str = "",
                 api_call: str = ""):
        self.timeseries_code = timeseries_code
        self.company = company
        self.check_name = check_name
        self.date = date  # filled in by error function
        self.value = value  # filled in by error function
        self.business_rule = business_rule  # filled in by error function
        self.reference = reference
        self.api_call = api_call

    def __repr__(self):
        return self.excel_format()

    def empty(self):
        if self.timeseries_code == "" and \
                self.company == "" and \
                self.check_name == "" and \
                self.value is None and \
                self.business_rule is None and \
                self.reference == "":
            return True
        return False

    def excel_format(self):
        lst = [self.timeseries_code, self.company, self.check_name, str(self.date), str(self.value), str(self.business_rule),
               self.reference, self.api_call]
        return ",".join(lst)


# metadata fields
from datetime import timedelta

from dateutil.relativedelta import relativedelta

id_string = "id"
calculated_string = "calculated"
frequency_string = "frequency"
chunks_string = "chunks"  # TODO: update after system update
min_value_string = "minValue"
max_value_string = "maxValue"
start_time_string = "startTime"
is_integer_string = "isInteger"
delay_string = "delay"

# frequencies
instant_string = "instant"
daily_string = "daily"
weekly_string = "weekly"
monthly_string = "monthly"
quarterly_string = "quarterly"
annually_string = "annually"

# Note: slight differences in days when using: 30, 90, etc. days versus 1 month, 3 months, and 13 weeks.
# 90 days ~ 3 months ~ 13 weeks (which is 91 days)
frequency_to_delta = {
    instant_string: relativedelta(seconds=+1),
    daily_string: relativedelta(days=+1),
    weekly_string: relativedelta(weeks=+1),  # unused
    monthly_string: relativedelta(months=+1),
    quarterly_string: relativedelta(months=+3),
    annually_string: relativedelta(years=+1),  # unused
}

frequency_to_timedelta = {
    instant_string: timedelta(seconds=+1),
    daily_string: timedelta(days=+1),
    weekly_string: timedelta(weeks=+1),  # unused
    monthly_string: timedelta(days=30),
    quarterly_string: timedelta(weeks=13),
    annually_string: timedelta(weeks=52),  # unused
}

frequencies = [k for k in frequency_to_delta]



import logging
from . import rule
from .timeseries_metadata import TimeseriesMetadataField
from .data_list import DataList
from .field import Field

class TimeseriesBusinessCheck(Field):
    def __init__(self):
        """
        :param metadata_names:
        :param rules: list of Rule objects
        """
        super().__init__()
        self.name = "Timeseries Check"
        self.value = {}

        self.data = DataList()
        self.fields = [self.data]  # will add fields when needed for use case
        self.rules = [rule.has_field(x.name) for x in self.fields]
        self.errors = {}

    def error(self, dct: dict):
        logging.error(dct)  # ERROR:root:error message
        for k, v in dct.items():
            if type(v) == str:
                v = [v]
            if k not in self.errors:
                self.errors[k] = v  # add new error list
            else:
                self.errors[k] = self.errors[k] + v  # append to old error list
        return

    def error_many(self, msgs: list):
        # msgs is a list of dictionaries
        if msgs:
            for msg in msgs:
                self.error(msg)
        return

    def run_check(self, new_check, condition=None):
        """
        Returns errors only from that check
        :param condition:
        :param new_check:
        :return:
        """
        error = new_check.run(self.data.value, condition)
        if error:
            self.error({new_check.name: error})
        return
